﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2019 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Problems.VehicleRouting.Variants;
using HeuristicLab.Random;
using HEAL.Attic;

namespace HeuristicLab.Problems.VehicleRouting.Encodings.Potvin {
  [Item("TemporalDistanceClusterCreator", "Creates a VRP solution by clustering customers first with a KMeans-algorithm based on their geographic position and building tours afterwards alternatevly in a random or a greedy fashion.")]
  [StorableType("8D2C7AEA-023B-4B6B-82D2-43FA8B0CA7FF")]
  public sealed class TemporalDistanceClusterCreator : ClusterCreator {

    [StorableConstructor]
    private TemporalDistanceClusterCreator(StorableConstructorFlag _) : base(_) { }

    public TemporalDistanceClusterCreator() : base() {
    }

    private TemporalDistanceClusterCreator(TemporalDistanceClusterCreator original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new TemporalDistanceClusterCreator(this, cloner);
    }

    public static PotvinEncoding CreateSolution(ITimeWindowedProblemInstance instance, IRandom random, int minK, int maxK, double clusterChangeThreshold, int creationOption) {
      PotvinEncoding result = new PotvinEncoding(instance);

      // (1) init data
      var times = new List<Tuple<double, double>>(instance.Cities.Value);
      var pdp = instance as IPickupAndDeliveryProblemInstance;
      for (int i = 1; i <= instance.Cities.Value; i++) {
        if (pdp == null || pdp.GetDemand(i) >= 0)
          times.Add(Tuple.Create(instance.ReadyTime[i], instance.DueTime[i]));
      }

      // (2) create a random number k of clusters
      int k = random.Next(minK, maxK);
      var kMeans = new KMeansAlgorithm<Tuple<double, double>>(CalculateMeanHelper, CalculateDistanceHelper);
      var clusters = kMeans.Run(times, k, clusterChangeThreshold, random);


      // (3) build tours with a (a) shuffling (b) greedy tour creation routine
      foreach (var cluster in clusters) {
        Tour newTour = new Tour();
        result.Tours.Add(newTour);

        if (creationOption == 0) {
          // (a) shuffle
          cluster.ShuffleInPlace(random);
          foreach (var customer in cluster) {
            newTour.Stops.Add(customer + 1);
          }
        } else {
          // (b) greedy
          foreach (var customer in cluster) {
            newTour.Stops.Add(customer + 1);
          }
          GreedyTourCreation(instance, result, newTour, true); // "true" means include costs
        }
      }

      return result;
    }

    private static Tuple<double, double> CalculateMeanHelper(List<Tuple<double, double>> times) {
      double meanReady = 0, meanDue = 0;
      foreach (var t in times) {
        meanReady += t.Item1 / times.Count;
        meanDue += t.Item2 / times.Count;
      }
      return Tuple.Create(meanReady, meanDue);
    }
    private static double CalculateDistanceHelper(Tuple<double, double> time1, Tuple<double, double> time2) {
      double distance = 0.0;
      distance += Math.Pow(time1.Item1 - time2.Item1, 2);
      distance += Math.Pow(time1.Item2 - time2.Item2, 2);
      return Math.Sqrt(distance);
    }

    public override IOperation InstrumentedApply() {
      IRandom random = RandomParameter.ActualValue;

      int minK = (MinK.Value.Value > 0) ? MinK.Value.Value : 1;
      int maxK = (MaxK.Value != null) ? MaxK.Value.Value : ProblemInstance.Vehicles.Value;
      double clusterChangeThreshold = (ClusterChangeThreshold.Value.Value >= 0.0 &&
                                       ClusterChangeThreshold.Value.Value <= 1.0)
        ? ClusterChangeThreshold.Value.Value
        : 0.0;

      // normalize probabilities
      double max = TourCreationProbabilities.Value.Max();
      double[] probabilites = new double[2];
      for (int i = 0; i < TourCreationProbabilities.Value.Length; i++) {
        probabilites[i] = TourCreationProbabilities.Value[i] / max;
      }

      List<int> creationOptions = new List<int>() { 0, 1 };
      int creationOption = creationOptions.SampleProportional(random, 1, probabilites, false, false).First();

      var instance = ProblemInstance as ITimeWindowedProblemInstance;
      if (instance == null) {
        throw new ArgumentException(string.Format("Cannot initialize {0} with data from {1}", instance.GetType(), ProblemInstance.GetType()));
      }

      VRPToursParameter.ActualValue = CreateSolution(instance, random, minK, maxK, clusterChangeThreshold, creationOption);
      return base.InstrumentedApply();
    }
  }
}
